# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Exploring MRIOs with Pymrio

# %% [markdown]
# The first step when working with a new MRIO data set is to familiarize yourself with the data.
# This notebook shows how to use the `pymrio` package to explore the data.
# We use the test data set that is included in the `pymrio` package.
# This is a completely artificial, very small MRIO.
# It is not meant to be realistic, but it is useful for developing, testing and learning.


# %% [markdown]
# First we import the required packages:

# %%
import pymrio

# %% [markdown]
# We can now load the test data set with the `load_test` function. We can call
# the MRIO whatever we want, here we use mrio.

# %%
mrio = pymrio.load_test()

# %% [markdown]
# We can get some first information about the MRIO by printing it.

# %%
print(mrio)

# %% [markdown]
# This tells us what the MRIO data we just loaded contains.
# We find a Z and Y matrix, some unit information and two satellite accounts, factor_inputs and emissions.

# %% [markdown]
# To get more specific data we can ask pymrio for regions, sectors, products, etc.

# %%
mrio.name

# %%
mrio.get_regions()

# %%
mrio.get_sectors()

# %%
mrio.get_Y_categories()

# %% [markdown]
# The same methods can be used to explore one of the satellite accounts.

# %%
print(mrio.emissions)

# %%
mrio.emissions.name
# %%
mrio.emissions.get_regions()


# %% [markdown]
# The satellite accounts also have a special method to get index (rows) of the acccounts.

# %%
mrio.emissions.get_rows()

# %% [markdown]
# # Searching through the MRIO

# %% [markdown]
# Several methods are available to search through the whole MRIO.
# These generally accept [regular expressions](https://docs.python.org/3/howto/regex.html) as search terms.

# %% [markdown]
# The most general method is 'find'. This can be used for a quick overview where a specific term appears in the MRIO.


# %%
mrio.find("air")

# %%
mrio.find("trade")

# %% [markdown]
# Not that 'find' (and all other search methods) a case sensitive.
# Do make a case insensitive search, add the regular expression flag `(?i)` to the search term.

# %%
mrio.find("value")

# %%
mrio.find("(?i)value")

# %% [markdown]
# ## Specific search methods: contains, match, fullmatch,

# %% [markdown]
# The MRIO class also contains a set of specific regular expresion search methods, mirroring the 'contains', 'match' and 'fullmatch'
# methods of the pandas DataFrame str column type. See the pandas documentation for details, in short:
#
#   - 'contains' looks for a match anywhere in the string
#   - 'match' looks for a match at the beginning of the string
#   - 'fullmatch' looks for a match of the whole string
#
# These methods are available for all index columns of the MRIO and have a similar signature:
#
#   1. As for 'find_all', the search term is case sensitive. To make it case insensitive, add the regular expression flag `(?i)` to the search term.
#   2. The search term can be passed to the keyword argument 'find_all' or as the first positional argument to search in all index levels.
#   3. Alternativels, the search term can be passed to the keyword argument with the level name to search only in that index level.
#
# The following examples show how to use these methods.

# %%
mrio.contains(find_all="ad")
mrio.contains("ad")  # find_all is the default argument

# %%
mrio.match("ad")

# %%
mrio.match("trad")

# %%
mrio.fullmatch("trad")

# %%
mrio.fullmatch("trade")

# %%
mrio.fullmatch("(?i).*AD.*")

# %% [markdown]
# For the rest of the notebook, we will do the examples with the 'contains' method, but the same applies to the other methods.

# %% [markdown]
# To search only at one specific level, pass the search term to the keyword argument with the level name.

# %%
mrio.contains(region="trade")

# %%
mrio.contains(sector="trade")

# %% [markdown]
# And of course, the method are also available for the satellite accounts.

# %%
mrio.emissions.contains(compartment="air")

# %% [markdown]
# Passing a non-existing level to the keyword argument is silently ignored.

# %%
mrio.factor_inputs.contains(compartment="trade")

# %% [markdown]
# This allows to search for terms that are only in some index levels.
# Logically, this is an 'or' search.
# %%
mrio.factor_inputs.contains(compartment="air", inputtype="Value")

# %% [markdown]
# But note, that if both levels exist, both must match (so it becomes a logical 'and').

# %%
mrio.emissions.contains(stressor="emission", compartment="air")

# %% [markdown]
# ## Search through all extensions

# %% [markdown]
# All three search methods are also available to loop through all extensions of the MRIO.

# %%
mrio.extension_contains(stressor="emission", compartment="air")

# %% [markdown]
# If only a subset of extensions should be searched, pass the extension names to the keyword argument 'extensions'.

# %% [markdown]
# ## Generic search method for any dataframe index

# %% [markdown]
# Internally, the class methods 'contains', 'match' and 'fullmatch' all the
# 'index_contains', 'index_match' and 'index_fullmatch' methods of ioutil module.
# This function can be used to search through index of any pandas DataFrame.

# %%
df = mrio.Y

# %% [markdown]
# Depending if a dataframe or an index is passed, the return is either the dataframe or the index.

# %%
pymrio.index_contains(df, "trade")

# %%
pymrio.index_contains(df.index, "trade")

# %%
pymrio.index_fullmatch(df, region="reg[2,4]", sector="m.*")

# %% [markdown]
# All search methods can easily be combined with the extract methods to extract the data that was found.
# For more information on this, see the [extract_data](./extract_data.ipynb) notebook.



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# # Characterization of stressors


# Stressor characterization allows the calculation of environmental and social impacts of economic activities. It transforms raw stressor data into meaningful impact indicators.

# Pymrio implements an innovative string-matching approach for characterization.
# This method matches stressors in the characterization table (provided in long format) with available stressors in the MRIO system. This brings the following benefits:

# - Ensures stressor correspondence across the MRIO system and characterization table
# - Performs automatic unit verification
# - Works regardless of entry order in the characterization table
# - Handles characterization tables that include factors for stressors not present in the satellite account
# - Efficiently manages region and sector-specific characterization factors
# - Enables characterization across different extensions

# This contrasts with traditional approaches that rely on matrix multiplication between stressor and characterization matrices, requiring strict 1:1 correspondence between matrix dimensions and precise ordering of entries.

# The characterization functionality is available both as an extension object method and as top-level function accepting complete MRIO objects or extension collections.

# In the following, we give some examples on how to use both methods, starting with some simple example and then advancing to more complex cases with regional specific factors.

# ## Basic Example

# For this example we use the test MRIO included in Pymrio. We also need
# the Pandas library for loading the characterization table and pathlib for some folder manipulation.


from pathlib import Path

import pandas as pd

import pymrio
from pymrio.core.constants import PYMRIO_PATH  # noqa

# To load the test MRIO we use:

io = pymrio.load_test()

# and the characterization table with some foo factors can be loaded by

charact_table = pd.read_csv(
    (PYMRIO_PATH["test_mrio"] / Path("concordance") / "emissions_charact.tsv"),
    sep="\t",
)
charact_table

# This table contains the columns 'stressor' and 'compartment' which correspond
# to the index names of the test_mrio emission satellite accounts:

io.emissions.F

# Theses index-names / columns-names need to match in order to match
# characterization factors to the stressors.

# The other columns names can be passed to the characterization method. By default the method assumes the following column names:
#
# - impact: name of the characterization/impact
# - factor: the numerical (float) multiplication value for a specific stressor to derive the impact/characterized account
# - impact_unit: the unit of the calculated characterization/impact
# - stressor_unit: the unit of the stressor in the extension
#
# Alternative names can be passed through the parameters
# *characterized_name_column*, *characterization_factors_column*, *characterized_unit_column* and *orig_unit_column*
#


# To calculate the characterization we use

char_emis = io.emissions.characterize(charact_table, name="impacts")

# The parameter *name* is optional, if omitted the name will be set to
# extension_name + _characterized. In case the passed name starts with an
# underscore, the return name with be the name of the original extension concatenated with the passed name.

# The return value is a named tuple with the *validation* and *extension* as attriubtes.

print(char_emis.extension)


char_emis.validation

# Checking the validation table is a recommended step that ensures accuracy and completeness before impact calculations. The validation process helps identify potential issues such as:
#
# - Missing characterization factors for specific region/sector/stressor combinations
# - Spelling mistakes or inconsistencies in stressor, sector, or region names
# - Unit mismatches between the MRIO system and characterization factors
# - Incomplete coverage that could affect impact assessment results
#
# By systematically checking these elements, users can avoid calculation errors and ensure their impact assessment captures all relevant environmental and social dimensions with the proper characterization factors.
#

# In the current case, the *charact_table* contains a characterization called 'total
# emissions', for which the calculation requires a stressor not present in the
# satellite account. This is indicated in the validation table in the *error_missing_stressor* column.
# The calculation can proceed, but for all impacts containing the stressor it is assumed to be 0.

# It is possible, to just the verification before doing any calculation with

only_val = io.emissions.characterize(charact_table, only_validation=True)
only_val.validation

# In that case the extension attribute is set to None.
# The same applies if a characterization needs to be aborted due to unit inconsistencies.

# Anyways, in case everything works as expected, the extension can be attached to the MRIO object.

io.impacts = char_emis.extension

# and used for subsequent calculations:

io.calc_all()
io.impacts.D_cba

# Note that units are checked against the unit specification of the extension.
# Thus, any mismatch of units will abort the calculation. The validation table
# helps to identify the issue.

charact_table.loc[charact_table.stressor == "emission_type1", "stressor_unit"] = "t"

ret_error = io.emissions.characterize(charact_table)

ret_error.extension

ret_error.validation

# The error_unit_impact column indicate the stressor with the unit mismatch.

# ## Regional specific characterization factors

# Here we use a table of regionally specific characterisation factors.
# The actual factors contained here are the same as in the basic example and we
# will modify them after loading.
# We will also investigate cases with missing data or conflicting units.
# The same principles can be used for sector specific characterization factors.

# We use the same data test mrio system as before:

io = pymrio.load_test()

# with the regional specific characterization factors from

charact_table_reg = pd.read_csv(
    (PYMRIO_PATH["test_mrio"] / Path("concordance") / "emissions_charact_reg_spec.tsv"),
    sep="\t",
)
charact_table_reg

# Compared with the previous table (charact_table), this table contains an additional
# column *region* which contains the regional specific data.
# Currently, the factors are actually the same as before, thus

char_reg = io.emissions.characterize(charact_table_reg)

# For regional specific characterization, the validation table contains information per region

char_reg.validation

# The extension is again available in the extension attribute

char_reg.extension.F

# gives the same result as before. To highlight regional
# specificity, we double the total emission factors of region 3.

charact_table_reg.loc[
    (charact_table_reg.region == "reg3")
    & (charact_table_reg.impact == "total emissions"),
    "factor",
] = (
    charact_table_reg.loc[
        (charact_table_reg.region == "reg3")
        & (charact_table_reg.impact == "total emissions"),
        "factor",
    ]
    * 2
)

# and calculate the new impacts

char_reg_dbl = io.emissions.characterize(charact_table_reg).extension
char_reg_dbl.F.loc["total emissions"]

# compared to

char_reg.extension.F.loc["total emissions"]

# ## Some more notes on validation


# We can put some more inconsistencies into the table to showcase the validation process.
# Some unit error in the stressors:

charact_table_reg.loc[
    (charact_table_reg.region == "reg4")
    & (charact_table_reg.stressor == "emission_type1"),
    "stressor_unit",
] = "s"

# Some inconsistent impact units:

charact_table_reg.loc[
    (charact_table_reg.region == "reg2")
    & (charact_table_reg.impact == "total emissions"),
    "impact_unit",
] = "kt"


# Some spelling mistake in region 2 for some stressor:

charact_table_reg.loc[
    (charact_table_reg.region == "reg2")
    & (charact_table_reg.stressor == "emission_type2"),
    "region",
] = "reg22"

# Another region data which is not available in the extension

new_data = charact_table_reg.iloc[[0]]
new_data.loc[:, "region"] = "reg_additional"
charact_table_reg = charact_table_reg.merge(new_data, how="outer")


report = io.emissions.characterize(charact_table_reg, only_validation=True).validation

# The unit errors are reported for each row, and the one additional region not present in the extension is report under *error_missing_region*.
# The column *error_unit_impact* indicates the impact with inconsistent units

report[report.stressor == "emission_type1"]

# In case of emission_type2, the *error_missing_region* is True for the whole stressor, since reg2 is "no longer present" in the factor sheets due to the spelling mistake.
# Thus, not all regions are covered in the specifications.
# Again, the column *error_unit_impact* indicates the impact with inconsistent units

report[report.stressor == "emission_type2"]

# ## Characterization across multiple extensions

# In addition to characterizing a single extension, pymrio also offers functionality
# to apply characterization across multiple extensions simultaneously. This is useful
# when your impacts depend on stressors that are distributed across different satellite accounts.

# Let's demonstrate this using our test MRIO system:

io = pymrio.load_test()

# First, let's create multiple extensions from our emissions data to better showcase this functionality:

# Create copies of the emissions extension with different names and data subsets
io.water = io.emissions.copy("water")
io.air = io.emissions.copy("air")

# Keep only water emissions in the water extension
io.water.F = io.water.F.loc[[("emission_type2", "water")], :]
io.water.F_Y = io.water.F_Y.loc[[("emission_type2", "water")], :]

# Keep only air emissions in the air extension
io.air.F = io.air.F.loc[[("emission_type1", "air")], :]
io.air.F_Y = io.air.F_Y.loc[[("emission_type1", "air")], :]

# Examining the extensions:

io.air.F


io.water.F

# To characterize across multiple extensions, we need a characterization table that includes
# an 'extension' column specifying which extension each stressor belongs to:

# Start with our regional characterization table
factors_reg_spec = pd.read_csv(
    (PYMRIO_PATH["test_mrio"] / Path("concordance") / "emissions_charact_reg_spec.tsv"),
    sep="\t",
)

# Create a copy and add an extension column based on compartment
factors_reg_ext = factors_reg_spec.copy()
factors_reg_ext.loc[:, "extension"] = factors_reg_ext.loc[:, "compartment"]

# Filter out any entries that don't correspond to our extensions
factors_reg_ext = factors_reg_ext[factors_reg_ext.compartment.isin(["air", "water"])]

# Examine our multi-extension characterization table:
factors_reg_ext.head(10)

# There are two ways to characterize across multiple extensions:

# 1. Using the top-level function with specific extensions:
ex_reg_multi = pymrio.extension_characterize(
    io.air,
    io.water,  # List the extensions you want to include
    factors=factors_reg_ext,
    new_extension_name="multi_top_level",
).extension

# 2. Using the MRIO object's method which automatically includes all available extensions:
ex_reg_mrio = io.extension_characterize(
    factors=factors_reg_ext, new_extension_name="multi_mrio_method"
).extension

# Both approaches produce the same result when the same extensions are involved:
print("Are the characterized F matrices equal?", ex_reg_multi.F.equals(ex_reg_mrio.F))

# Add the extension to our MRIO and calculate results:
io.multi = ex_reg_multi
io.calc_all()
io.multi.D_cba

# As with single extension characterization, validation is crucial:
validation_report = pymrio.extension_characterize(
    io.air, io.water, factors=factors_reg_ext, only_validation=True
).validation

print("Validation report:")
validation_report

# The validation process helps identify issues such as:
# - Missing stressors or extensions
# - Unit inconsistencies
# - Missing regions or sectors
# - Extension name mismatches

# Important considerations for multi-extension characterization:
#
# 1. The 'extension' column in your characterization table must match the extension names in your MRIO
# 2. All extensions must have compatible region and sector classifications
# 3. Units must be consistent across extensions and characterization factors
# 4. If a characterization table references an extension that doesn't exist,
#    it will be noted in the validation report



"""Meta data for provenance and version tracking in pymrio."""

import datetime
import getpass
import json
import logging
import os
import platform
import zipfile
from collections import OrderedDict
from pathlib import Path

from pymrio.core.constants import DEFAULT_FILE_NAMES
from pymrio.version import __version__ as pymrio_version


class MRIOMetaData:
    """MRIO Meta Data class."""

    def __init__(
        self,
        location=None,
        description=None,
        name=None,
        system=None,
        version=None,
        path_in_arc="",
        logger_function=logging.info,
    ):
        """Organzise the MRIO meta data.

        The meta data is stored in a json file.

        Note:
        -----
            The parameters 'description', 'name', 'system', and
            'version' should be set during the establishment of the meta data
            file.  If the meta data file already exists and they are given
            again, the corresponding entry will be overwritten if
            replace_existing_meta_content is set to True (with a note in
            the 'History' field.)

        Parameters
        ----------
        location: str, valid path, optional
            Path or file for loading a previously saved metadata file and/or
            saving additional metadata (the method 'save' will use this
            location by default).  This can be the full file path or just the
            storage folder.  In the latter case, the filename defined in
            DEFAULT_FILE_NAMES['metadata'] (currently 'metadata.json') is
            assumed. The metadata can also be read from an archived zip file.
            In this case the data is 'read only' and can not be stored back
            to the same archive directly (writing data as zip archive not
            implemented yet). This means, that a new file has to be specified
            when saving the metadata.

        description: str, optional
            Description of the metadata file purpose and mrio,
            default set to 'Metadata file for pymrio'.
            Will be set the first time the metadata file gets established;
            subsequent changes are recorded in 'history'.

        name: str, optional
            Name of the mrio (e.g. wiod, exiobase)
            Will be set the first time the metadata file gets established;
            subsequent changes are recorded in 'history'.

        system: str, optional
            For example 'industry by industry', 'ixi', ...
            Will be set the first time the metadata file gets established;
            subsequent changes are recorded in 'history'.

        version: str, int, float, optional
            Version number
            Will be set the first time the metadata file gets established;
            subsequent changes are recorded in 'history'.

        path_in_arc: string, optional
            Path to the meta data in a zip file. path_in_arc must be given
            without leading dot or slash; thus to point to the data in the
            root of the compressed file pass '', for data in e.g. the folder
            'emissions' pass 'emissions/'.  Only used if parameter 'location'
            points to a compressed zip file.

        logger_function: func, optional
            Function accepting strings.
            The info string written to the metadata is also
            passed to this function. By default, the funtion
            is set to logging.info. Set to None for no output.

        """
        if location:
            location = Path(location)
            self._path_in_arc = None
            if location.is_file():
                self._metadata_file = location
                if zipfile.is_zipfile(str(location)):
                    with zipfile.ZipFile(file=str(location)) as zf:
                        if path_in_arc not in zf.namelist():
                            path_in_arc = os.path.join(path_in_arc, DEFAULT_FILE_NAMES["metadata"])

                    self._path_in_arc = str(path_in_arc)

            elif location.is_dir():
                self._metadata_file = location / DEFAULT_FILE_NAMES["metadata"]
            else:
                if os.path.splitext(str(location))[1] == "":
                    self._metadata_file = location / DEFAULT_FILE_NAMES["metadata"]
                else:
                    self._metadata_file = location
        else:
            self._metadata_file = None

        self.logger = logger_function if logger_function else lambda x: None

        if self._metadata_file and self._metadata_file.is_file():
            self._read_content()
            self.logger(f"Read metadata from {self._metadata_file}")

            if zipfile.is_zipfile(str(self._metadata_file)):
                self._metadata_file = None
                self._path_in_arc = None

            if description:
                self.change_meta("description", description)
            if name:
                self.change_meta("name", name)
            if system:
                self.change_meta("system", system)
            if version:
                self.change_meta("version", version)

        else:
            if not description:
                description = "Metadata for pymrio"
            self._content = OrderedDict(
                [
                    ("description", description),
                    ("name", name),
                    ("system", system),
                    ("version", version),
                    ("history", []),
                ]
            )
            self.logger("Start recording metadata")

    def __repr__(self):
        """Repr."""
        return self.__str__()

    def __str__(self):
        """Str."""
        nr_hist_lines_show = 10

        hist = "\n".join(self._content["history"][:nr_hist_lines_show])

        if nr_hist_lines_show < len(self._content["history"]):
            hist = hist + "\n ... (more lines in history)"

        return (
            f"Description: {self.description}\n"
            f"MRIO Name: {self.name}\n"
            f"System: {self.system}\n"
            f"Version: {self.version}\n"
            f"File: {self._metadata_file}\n"
            f"History:\n{hist}"
        )

    def _make_download_log(
        location,
        name,
        description=None,
        system=None,
        version=None,
        logger_function=logging.info,
    ):
        """Make a download log file (factory method).

        This makes an instance of the metahander suitable for storing download information.
        During estabishment it also records system information (username, hostname, pymrio version, etc.)
        as defined in _get_system_meta.

        Parameters
        ----------
        location: str, valid path
            Path or filename for the history log file.
            This can be the full file path or just the
            storage folder.  In the latter case, the filename defined in
            DEFAULT_FILE_NAMES['download_log'] (currently 'download_log.json') is
            assumed.

        name: str
            Name of the mrio (e.g. wiod, exiobase)
            Will be set the first time the metadata file gets established;
            subsequent changes are recorded in 'history'.

        description: str, optional
            Description of the metadata file purpose and mrio,
            default set to 'Metadata file for pymrio'.
            Will be set the first time the metadata file gets established;
            subsequent changes are recorded in 'history'.

        system: str, optional
            For example 'industry by industry', 'ixi', ...
            Will be set the first time the metadata file gets established;
            subsequent changes are recorded in 'history'.

        version: str, int, float, optional
            Version number
            Will be set the first time the metadata file gets established;
            subsequent changes are recorded in 'history'.

        logger_function: func, optional
            Function accepting strings.
            The info string written to the metadata is also
            passed to this function. By default, the funtion
            is set to logging.info. Set to None for no output.

        Returns
        -------
        MRIOMetaData object
            setup as a download log

        """
        location = Path(location)

        if location.exists():
            if location.is_file():
                full_location = location
            elif location.is_dir():
                full_location = location / DEFAULT_FILE_NAMES["download_log"]
            else:
                raise ValueError("Location must be a valid file or directory")
        else:
            if location.suffix == "":
                full_location = location / DEFAULT_FILE_NAMES["download_log"]
            else:
                full_location = location

        mmd = MRIOMetaData(
            location=full_location,
            description=description,
            name=name,
            system=system,
            version=version,
            logger_function=logger_function,
        )

        for typ, val in mmd._get_system_meta().items():
            mmd.note(f"{typ}: {val}")

        mmd.note("Download log created")

        return mmd

    def _read_content(self):
        """Read the metadata file and store the content in self._content."""
        if self._metadata_file:
            if self._path_in_arc:
                with zipfile.ZipFile(file=str(self._metadata_file)) as zf:
                    with zf.open(self._path_in_arc) as f:
                        self._content = json.load(f, object_pairs_hook=OrderedDict)
            else:
                with open(str(self._metadata_file)) as f:
                    self._content = json.load(f, object_pairs_hook=OrderedDict)

    def note(self, entry):
        """Add the passed string as note to the history.

        If log is True (default), also log the string by logging.info
        """
        self._add_history(entry_type="NOTE", entry=entry)

    def _add_fileio(self, entry):
        """Add the passed string as FILEIO to the history."""
        self._add_history(entry_type="FILEIO", entry=entry)

    def _add_modify(self, entry):
        """Add the passed string as MODIFICATION to the history."""
        self._add_history(entry_type="MODIFICATION", entry=entry)

    def _add_history(self, entry_type, entry):
        """Add entry as entry_type to the history."""
        meta_string = f"{self._time()} - {entry_type.upper()} -  {entry}"

        self._content["history"].insert(0, meta_string)
        self.logger(meta_string)

    @property
    def metadata(self):
        """Return the current content."""
        return self._content

    @property
    def history(self):
        """Return all recorded history."""
        return self._content["history"]

    @property
    def modification_history(self):
        """Return all modification history entries."""
        return self._get_history_type("MODIFICATION")

    @property
    def note_history(self):
        """Return all note history entries."""
        return self._get_history_type("NOTE")

    @property
    def file_io_history(self):
        """Return all fileio history entries."""
        return self._get_history_type("FILEIO")

    def _get_history_type(self, history_type):
        return [ent for ent in self._content["history"] if history_type.upper() in ent]

    @property
    def description(self):
        """Return description entries."""
        return self._content["description"]

    @property
    def name(self):
        """Return the name entry."""
        return self._content["name"]

    @property
    def system(self):
        """Return the system entry."""
        return self._content["system"]

    @property
    def version(self):
        """Return version."""
        return self._content["version"]

    def change_meta(self, para, new_value, log=True):
        """Change the meta data.

        This function does nothing if None is passed as new_value.
        To set a certain value to None pass the str 'None'

        Parameters
        ----------
        para: str
            Meta data entry to change

        new_value: str
            New value

        log: boolean, optional
            If True (default) records the meta data change
            in the history

        """
        if not new_value:
            return
        para = para.lower()
        if para == "history":
            raise ValueError('History can only be extended - use method "note"')
        old_value = self._content.get(para, None)
        if new_value == old_value:
            return
        self._content[para] = new_value
        if old_value and log:
            self._add_history(
                entry_type="METADATA_CHANGE",
                entry=f'Changed parameter "{para}" from "{old_value}" to "{new_value}"',
            )

    def _get_system_meta(self):
        """Return a dictionary with user meta data.

        with username, computername, os, os version, and python version
        """
        try:
            username = getpass.getuser()
        except Exception:
            username = "unknown"

        try:
            hostname = platform.node()
        except Exception:
            hostname = "unknown"

        try:
            os = platform.system()
        except Exception:
            os = "unknown"

        try:
            pyver = platform.python_version()
        except Exception:
            pyver = "unknown"

        return {
            "username": username,
            "hostname": hostname,
            "os": os,
            "pymrio_version": pymrio_version,
            "python_version": pyver,
        }

    def _time(self):
        return f"{datetime.datetime.now():%Y%m%d %H:%M:%S}"

    def _read_content(self):
        """Read metadata from location (and path_in_arc if archive).

        This function is called during the init process and
        should not be used in isolation: it overwrites
        unsafed metadata.
        """
        if self._path_in_arc:
            with zipfile.ZipFile(file=str(self._metadata_file)) as zf:
                self._content = json.loads(
                    zf.read(self._path_in_arc).decode("utf-8"),
                    object_pairs_hook=OrderedDict,
                )
        else:
            with self._metadata_file.open("r") as mdf:
                self._content = json.load(mdf, object_pairs_hook=OrderedDict)

    def save(self, location=None):
        """Save the current status of the metadata.

        This saves the metadata at the location of the previously loaded
        metadata or at the file/path given in location.

        Specify a location if the metadata should be stored in a different
        location or was never stored before. Subsequent saves will use the
        location set here.

        Parameters
        ----------
        location: str, optional
            Path or file for saving the metadata.
            This can be the full file path or just the storage folder.
            In the latter case, the filename defined in
            DEFAULT_FILE_NAMES['metadata'] (currently 'metadata.json') is
            assumed.
        """
        if location:
            location = Path(location)
            if os.path.splitext(str(location))[1] == "":
                self._metadata_file = location / DEFAULT_FILE_NAMES["metadata"]
            else:
                self._metadata_file = location

        if self._metadata_file:
            with self._metadata_file.open(mode="w") as mdf:
                json.dump(self._content, mdf, indent=4)
        else:
            logging.error("No metadata file given for storing the file")

    def __call__(self, note=None):
        """Call (shortcut) for showing and adding notes.

        If called without parameters, prints a full description
        of the metadata. If 'note' is given, add note and prints
        content.
        """
        if note:
            self.note(entry=note)

        print(self.__str__())



"""Mathematical functions for input output calculations.

All methods here should follow the functional programming paradigm

Note:
----
To avoid namespace pollution everything here starts with calc_

"""

import typing
import warnings
from collections import namedtuple

import numpy as np
import pandas as pd

import pymrio.tools.ioutil as ioutil


def calc_x(Z, Y):
    """Calculate the industry output x from the Z and Y matrix.

    industry output (x) = flows (sum_columns(Z)) + final demand (sum_columns(Y))

    Parameters
    ----------
    Z : pandas.DataFrame or numpy.array
        Symmetric input output table (flows)
    Y : pandas.DataFrame or numpy.array
        final demand with categories (1.order) for each country (2.order)

    Returns
    -------
    pandas.DataFrame or numpy.array
        Industry output x as column vector
        The type is determined by the type of Z. If DataFrame index as Z

    """
    x = np.reshape(np.sum(np.hstack((Z, Y)), 1), (-1, 1))
    if type(Z) is pd.DataFrame:
        x = pd.DataFrame(x, index=Z.index, columns=["indout"])
    if type(x) is pd.Series:
        x = pd.DataFrame(x)
    if type(x) is pd.DataFrame:
        x.columns = ["indout"]
    return x


def calc_x_from_L(L, y):
    """Calculate the industry output x from L and a y vector.

    x = Ly

    The industry output x is computed from a demand vector y

    Parameters
    ----------
    L : pandas.DataFrame or numpy.array
        Symmetric input output Leontief table
    y : pandas.DataFrame or numpy.array
        a column vector of the total final demand

    Returns
    -------
    pandas.DataFrame or numpy.array
        Industry output x as column vector
        The type is determined by the type of L. If DataFrame index as L

    """
    x = L.dot(y)
    if type(x) is pd.Series:
        x = pd.DataFrame(x)
    if type(x) is pd.DataFrame:
        x.columns = ["indout"]
    return x


def calc_Z(A, x):
    """Calculate the Z matrix (flows) from A and x.

    A = Z / x[None, :]  =>  Z = A * x[None, :]

    By definition, the coefficient matrix A is basically the normalized flows
    So Z is just derived from A by un-normalizing using the industrial output x

    Parameters
    ----------
    A : pandas.DataFrame or numpy.array
        Symmetric input output table (coefficients)
    x : pandas.DataFrame or numpy.array
        Industry output column vector

    Returns
    -------
    pandas.DataFrame or numpy.array
        Symmetric input output table (flows) Z
        The type is determined by the type of A.
        If DataFrame index/columns as A

    """
    if (type(x) is pd.DataFrame) or (type(x) is pd.Series):
        x = x.to_numpy()
    x = x.reshape((1, -1))  # use numpy broadcasting - much faster
    # (but has to ensure that x is a row vector)
    # old mathematical form:
    # return A.dot(np.diagflat(x))
    if type(A) is pd.DataFrame:
        return pd.DataFrame(A.to_numpy() * x, index=A.index, columns=A.columns)
    return A * x


def calc_A(Z, x):
    """Calculate the A matrix (coefficients) from Z and x.

    A is a normalized version of the industrial flows Z

    Parameters
    ----------
    Z : pandas.DataFrame or numpy.array
        Symmetric input output table (flows)
    x : pandas.DataFrame or numpy.array
        Industry output column vector

    Returns
    -------
    pandas.DataFrame or numpy.array
        Symmetric input output table (coefficients) A
        The type is determined by the type of Z.
        If DataFrame index/columns as Z

    """
    if (type(x) is pd.DataFrame) or (type(x) is pd.Series):
        x = x.to_numpy()
    if (type(x) is not np.ndarray) and (x == 0):
        recix = 0
    else:
        with warnings.catch_warnings():
            # catch the divide by zero warning
            # we deal wit that by setting to 0 afterwards
            warnings.simplefilter("ignore")
            recix = 1 / x
        recix[recix == np.inf] = 0
        recix = recix.reshape((1, -1))
    # use numpy broadcasting - factor ten faster
    # Mathematical form - slow
    # return Z.dot(np.diagflat(recix))
    if type(Z) is pd.DataFrame:
        return pd.DataFrame(Z.to_numpy() * recix, index=Z.index, columns=Z.columns)
    return Z * recix


def calc_B(Z, x):
    """Calculate the B matrix (coefficients) from Z and x.

    B is a normalized version of the industrial flows of the
    transpose of Z, which quantifies the input to downstream
    sectors.

    Parameters
    ----------
    Z : pandas.DataFrame or numpy.array
        Symmetric input output table (flows)
    x : pandas.DataFrame or numpy.array
        Industry output column vector

    Returns
    -------
    pandas.DataFrame or numpy.array
        Symmetric input output table (coefficients) As
        The type is determined by the type of Z.
        If DataFrame index/columns as Z

    """
    if (type(x) is pd.DataFrame) or (type(x) is pd.Series):
        x = x.to_numpy()
    if (type(x) is not np.ndarray) and (x == 0):
        recix = 0
    else:
        with warnings.catch_warnings():
            # catch the divide by zero warning
            # we deal wit that by setting to 0 afterwards
            warnings.simplefilter("ignore")
            recix = 1 / x
        recix[recix == np.inf] = 0
        recix = recix.reshape((1, -1))
    # use numpy broadcasting - factor ten faster
    # Mathematical form - slow
    # return np.diagflat(recix).dot(Z)
    if type(Z) is pd.DataFrame:
        return pd.DataFrame(
            np.transpose(np.transpose(Z.values) * recix),
            index=Z.index,
            columns=Z.columns,
        )
    return np.transpose(np.transpose(Z) * recix)


def calc_L(A):
    """Calculate the Leontief L from A.

    L = inverse matrix of (I - A)

    Where I is an identity matrix of same shape as A

    Comes from:
        x = Ax + y  =>  (I-A)x = y
    Where:
        A: coefficient input () - output () table
        x: output vector
        y: final demand vector

    Hence, L allows to derive a required output vector x for a given demand y

    Parameters
    ----------
    A : pandas.DataFrame or numpy.array
        Symmetric input output table (coefficients)

    Returns
    -------
    pandas.DataFrame or numpy.array
        Leontief input output table L
        The type is determined by the type of A.
        If DataFrame index/columns as A

    """
    I = np.eye(A.shape[0])  # noqa
    if type(A) is pd.DataFrame:
        return pd.DataFrame(np.linalg.inv(I - A), index=A.index, columns=A.columns)
    return np.linalg.inv(I - A)


def calc_G(B, L=None, x=None):
    """Calculate the Ghosh inverse matrix G.

    Either from B (high computation effort) or from Leontief
    matrix L and x (low computation effort).

    G = inverse matrix of (I - B) = hat(x) *  L * hat(x)^{-1}

    where I is an identity matrix of same shape as B, and hat(x) is the diagonal
    matrix with values of x on the diagonal.

    Parameters
    ----------
    B : pandas.DataFrame or numpy.array
        Symmetric input output table (coefficients)

    Returns
    -------
    pandas.DataFrame or numpy.array
        Ghosh input output table G
        The type is determined by the type of B.
        If DataFrame index/columns as B

    """
    # if L has already been calculated, then G can be derived from it
    # with low computational cost.
    if L is not None and x is not None:
        if (type(x) is pd.DataFrame) or (type(x) is pd.Series):
            x = x.to_numpy()
        if (type(x) is not np.ndarray) and (x == 0):
            recix = 0
        else:
            with warnings.catch_warnings():
                # catch the divide by zero warning
                # we deal wit that by setting to 0 afterwards
                warnings.simplefilter("ignore")
                recix = 1 / x
            recix[recix == np.inf] = 0
            recix = recix.reshape((1, -1))

        if type(L) is pd.DataFrame:
            return pd.DataFrame(
                np.transpose(recix * np.transpose(L.to_numpy() * x)),
                index=L.index,
                columns=L.columns,
            )
        # G = hat(x) *  L * hat(x)^{-1} in
        # mathematical form hatx.dot(L.transpose()).dot(np.linalg.inv(hatx)).
        # it is computationally much faster to multiply
        # element-wise because hatx is a diagonal matrix.
        return np.transpose(recix * np.transpose(L * x))
    # calculation of the inverse of I-As has a high computational cost.
    I = np.eye(B.shape[0])  # noqa
    if type(B) is pd.DataFrame:
        return pd.DataFrame(np.linalg.inv(I - B), index=B.index, columns=B.columns)
    return np.linalg.inv(I - B)  # G = inverse matrix of (I - B)


def calc_S(F, x):
    """Calculate extensions/factor inputs coefficients.

    Parameters
    ----------
    F : pandas.DataFrame or numpy.array
        Total direct impacts
    x : pandas.DataFrame or numpy.array
        Industry output column vector

    Returns
    -------
    pandas.DataFrame or numpy.array
        Direct impact coefficients S
        The type is determined by the type of F.
        If DataFrame index/columns as F

    """
    return calc_A(F, x)


def calc_S_Y(F_Y, y):
    """Calculate extensions/factor inputs coefficients for the final demand.

    Note:
    ----
    F_Y will be restricted to the item available in y for the calculation. This
    allows to use a subset of Y (just some regions for example) to be used for
    the calculations. Only works when using pandas DataFrames/Series.


    Parameters
    ----------
    F_Y : pandas.DataFrame or numpy.array
         Final demand impacts
    y : pandas.DataFrame or numpy.array
        Final demand column vector (e.g. mrio.Y.sum())

    Returns
    -------
    pandas.DataFrame or numpy.array
        Direct impact coefficients of final demand S_Y
        The type is determined by the type of F.
        If DataFrame index/columns as F

    """
    if type(F_Y) is pd.DataFrame and type(y) is (pd.Series or pd.DataFrame):
        F_Y_calc = F_Y.loc[:, y.index]
    else:
        F_Y_calc = F_Y
    return calc_A(F_Y_calc, y)


def calc_F(S, x):
    """Calculate total direct impacts from the impact coefficients.

    Parameters
    ----------
    S : pandas.DataFrame or numpy.array
        Direct impact coefficients S
    x : pandas.DataFrame or numpy.array
        Industry output column vector

    Returns
    -------
    pandas.DataFrame or numpy.array
        Total direct impacts F
        The type is determined by the type of S.
        If DataFrame index/columns as S

    """
    return calc_Z(S, x)


def calc_F_Y(S_Y, y):
    """Calc total direct impacts from the impact coefficients of final demand.

    Note:
    ----
    S_Y will be restricted to the item available in y for the calculation. This
    allows to use a subset of Y (just some regions for example) to be used for
    the calculations. Only works when using pandas DataFrames/Series.

    Parameters
    ----------
    S_Y : pandas.DataFrame or numpy.array
        Direct impact coefficients of final demand
    y : pandas.DataFrame or numpy.array
        Final demand column vector (e.g. mrio.Y.sum())

    Returns
    -------
    pandas.DataFrame or numpy.array
        Total direct impacts of final demand F_Y
        The type is determined by the type of S_Y.
        If DataFrame index/columns as S_Y

    """
    if type(S_Y) is pd.DataFrame and type(y) is (pd.Series or pd.DataFrame):
        S_Y_calc = S_Y.loc[:, y.index]
    else:
        S_Y_calc = S_Y

    return calc_Z(S_Y_calc, y)


def calc_M(S, L):
    """Calculate multipliers of the extensions.

    Parameters
    ----------
    L : pandas.DataFrame or numpy.array
        Leontief input output table L
    S : pandas.DataFrame or numpy.array
        Direct impact coefficients

    Returns
    -------
    pandas.DataFrame or numpy.array
        Multipliers M
        The type is determined by the type of S.
        If DataFrame index/columns as S

    """
    return S.dot(L)


def calc_M_down(S, G):
    """Calculate downstream multipliers of the extensions.

    M_down = S * ( G^T - I )

    Where I is an identity matrix of same shape as G

    Parameters
    ----------
    G : pandas.DataFrame or numpy.array
        Ghosh input output table G
    S : pandas.DataFrame or numpy.array
        Direct impact coefficients

    Returns
    -------
    pandas.DataFrame or numpy.array
        Downstream multipliers M
        The type is determined by the type of S.
        If DataFrame index/columns as S

    """
    return S.dot(np.transpose(G) - np.eye(G.shape[0]))


def calc_e(M, Y):
    """Calculate total impacts (footprints of consumption Y).

    Parameters
    ----------
    M : pandas.DataFrame or numpy.array
        Multipliers
    Y : pandas.DataFrame or numpy.array
        Final consumption

    TODO - this must be completely redone (D, check for dataframe, ...)

    Returns
    -------
    pandas.DataFrame or numpy.array
        Multipliers m
        The type is determined by the type of M.
        If DataFrame index/columns as M
        The calculation is based on multipliers M and final demand Y

    """
    return M.dot(Y)


def recalc_M(S, D_cba, Y):
    """Calculate Multipliers based on footprints.

    Parameters
    ----------
    D_cba : pandas.DataFrame or numpy array
        Footprint per sector and country
    Y : pandas.DataFrame or numpy array
        Final demand: aggregated across categories or just one category, one
        column per country. This will be diagonalized per country block.
        The diagonolized form must be invertable for this method to work.

    Returns
    -------
    pandas.DataFrame or numpy.array
        Multipliers M
        The type is determined by the type of D_cba.
        If DataFrame index/columns as D_cba
    """
    Y_diag = ioutil.diagonalize_columns_to_sectors(Y)
    Y_inv = np.linalg.inv(Y_diag)
    M = D_cba.dot(Y_inv)
    if type(D_cba) is pd.DataFrame:
        M.columns = D_cba.columns
        M.index = D_cba.index

    return M


def calc_accounts(S, L, Y):
    """Calculate sector specific cba and pba based accounts, imp and exp accounts.

    The total industry output x for the calculation
    is recalculated from L and y

    Parameters
    ----------
    L : pandas.DataFrame
        Leontief input output table L
    S : pandas.DataFrame
        Direct impact coefficients
    Y : pandas.DataFrame
        Final demand: aggregated across categories or just one category, one
        column per country. Shape: rows as L, Y with just region names

    Returns
    -------
    Tuple
        (D_cba, D_pba, D_imp, D_exp)

        Format: D_row x L_col (=nr_countries*nr_sectors)

        - D_cba        Footprint per sector and country
        - D_pba      Total factur use per sector and country
        - D_imp       Total global factor use to satisfy total final demand in
                      the country per sector
        - D_exp       Total factor use in one country to satisfy final demand
                      in all other countries (per sector)
    """
    # diagonalize each sector block per country
    # this results in a disaggregated y with final demand per country per
    # sector in one column

    if isinstance(Y.columns, pd.MultiIndex):
        raise ValueError("Column index of Y can not be a MultiIndex - aggregate the columns")
    Y_diag = ioutil.diagonalize_columns_to_sectors(Y)
    x_diag = L @ Y_diag

    x_tot = x_diag.to_numpy().sum(1)

    del Y_diag

    D_cba = pd.DataFrame(S @ x_diag)

    # D_pba = S.dot(np.diagflat(x_tot))
    # faster broadcasted calculation:
    # NB: D_pba columns might be different to D_cba columns if Y include "regions"
    # which are not in the core. This happens for example for statistical discrepancy
    # in the OECD tables. It is "theoretically" possible to calculate footprints for
    # these "regions", but not PBA accounts.
    D_pba = pd.DataFrame(S.to_numpy() * x_tot.reshape((1, -1)), index=S.index, columns=S.columns)

    # for the traded accounts set the domestic industry output to zero
    x_trade = ioutil.set_dom_block(x_diag, value=0)
    D_imp = pd.DataFrame(S @ x_trade)

    x_exp = x_trade.sum(1).to_numpy()

    # D_exp = S.dot(np.diagflat(x_exp))
    # faster broadcasted version:
    D_exp = pd.DataFrame(S.to_numpy() * x_exp.reshape((1, -1)), index=S.index, columns=S.columns)

    return (D_cba, D_pba, D_imp, D_exp)


def calc_gross_trade(Z: pd.DataFrame, Y: pd.DataFrame) -> typing.NamedTuple(
    "gross_trade", [("bilat_flows", pd.DataFrame), ("totals", pd.DataFrame)]
):
    """Calculate the gross bilateral trade flows and totals.

    These are the entries of Z and Y with the domestic blocks set to 0.

    Notes
    -----
    This only works for DataFrame representation of Z and Y following the
    standard pymrio structure (regions on Multiindex level 0 or named 'region',
    sectors/categories on Multiindex level 1 or named 'sector').

    Parameters
    ----------
    Z : pandas.DataFrame
        Symmetric input output table (flows)
    Y : pandas.DataFrame
        final demand with regions (multiindex level 1) and categories (level 2)

    Returns
    -------
    namedtuple (with two DataFrames)
        A NamedTuple with two fields:

            - bilat_flows: df with rows: exporting country and sector,
              columns: importing countries
            - totals: df with gross total imports and exports per sector
              and region

    """
    Z_trade_blocks = ioutil.set_dom_block(Z, value=0)
    Y_trade_blocks = ioutil.set_dom_block(Y, value=0)

    level_spec_Z = "region" if "region" in Z.columns.names else 0
    level_spec_Y = "region" if "region" in Y.columns.names else 0
    Z_trade_agg = Z_trade_blocks.T.groupby(level=level_spec_Z, sort=False).agg("sum").T
    Y_trade_agg = Y_trade_blocks.T.groupby(level=level_spec_Y, sort=False).agg("sum").T

    x_bilat = Z_trade_agg + Y_trade_agg

    level_spec_x = "sector" if "sector" in x_bilat.index.names else 1
    gross_imports = pd.DataFrame(
        x_bilat.groupby(level=level_spec_x, sort=False).agg("sum").stack().swaplevel(),
        columns=["imports"],
    )
    gross_exports = pd.DataFrame(x_bilat.sum(axis=1), columns=["exports"])

    # gross_exports first as the swaplevel in gross_imports rearanges sectors!
    gross_totals = gross_exports.join(gross_imports)

    return namedtuple("gross_trade", "bilat_flows totals")(x_bilat, gross_totals)



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.2
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Extract data from Pymrio

# %% [markdown]
# This notebook shows how to extract specific data from the pymrio object for further processing in Python. For exporting/saving the data to another file format see [the notebook on saving/loading/exporting data.](./load_save_export.ipynb)

# %%
import pymrio

# %%
mrio = pymrio.load_test().calc_all()

### Basic pandas indexing of pymrio tables

# %% [markdown]
# Since pymrio is built on top of pandas, we can use the pandas functions to extract data from the pymrio object. For example, to access the part of the A matrix from the region 2 we can use:

# %%
A_reg2 = mrio.A.loc["reg2", "reg2"]
A_reg2

# %% [markdown]
# Most tables are indexed via a multiindex, in case of the A matrix the index is a tuple of the region and the sector.
# To access all technical coefficients (column) data for mining from all regions we can use:


# %%
A_mining = mrio.A.loc[:, (slice(None), "mining")]
A_mining

# %% [markdown]
# For further information on the pandas multiindex see the [pandas documentation on advanced indexing.](https://pandas.pydata.org/docs/user_guide/advanced.html)
#
# ## Extracting data across extension tables

# %% [markdown]
# Pymrio includes methods for bulk extraction of data across extension tables. These can either work on a specific extension or across all extensions of the system.

# %% [markdown]
# ### Extracting from a specific extension


# %% [markdown]
# Here we use use the `extract` method available in the extension object.
# This expect a list of rows (index) to extract.
# Here we extract some rows from the emission extension table.
# To do so, we first define the rows (index) to extract:

# %%
rows_to_extract = [("emission_type1", "air"), ("emission_type2", "water")]

# %% [markdown]
# We can now use the `extract` method to extract the data, either as a pandas DataFrame

# %%
df_extract = mrio.emissions.extract(rows_to_extract, return_type="dataframe")
df_extract.keys()

# %% [markdown]
# Or we extract into a new extension object:

# %%
ext_extract = mrio.emissions.extract(rows_to_extract, return_type="extension")
str(ext_extract)

# %% [markdown]
# Note that the name of the extension object is now `Emissions_extracted`, based on the name of the original extension object.
# To use another name, just pass the name as the `return_type` method.

# %%
new_extension = mrio.emissions.extract(rows_to_extract, return_type="new_extension")
str(new_extension)

# %% [markdown]
# Extracting to dataframes is also a convienient
# way to convert an extension object to a dictionary:

# %%
df_all = mrio.emissions.extract(mrio.emissions.get_rows(), return_type="dfs")
df_all.keys()


# The method also allows to only extract some of the accounts:
df_some = mrio.emissions.extract(
    mrio.emissions.get_rows(), dataframes=["D_cba", "D_pba"], return_type="dfs"
)
df_some.keys()


# %% [markdown]
# ### Extracting from all extensions

# %% [markdown]
# We can also extract data from all extensions at once.
# This is done using the `extension_extract` method from the pymrio object.
# This expect a dict with keys based on the extension names and values as a list of rows (index) to extract.

# %% [markdown]
# Lets assume we want to extract value added and all emissions.
# We first define the rows (index) to extract:

# %%
to_extract = {
    "Factor Inputs": "Value Added",
    "Emissions": [("emission_type1", "air"), ("emission_type2", "water")],
}


# %% [markdown]
# And can then use the `extension_extract` method to extract the data, either as a pandas DataFrame,
# which returns a dictionary with the extension names as keys

# %%
df_extract_all = mrio.extension_extract(to_extract, return_type="dataframe")
df_extract_all.keys()

# %%
df_extract_all["Factor Inputs"].keys()

# %% [markdown]
# We can also extract into a dictionary of extension objects:

# %%
ext_extract_all = mrio.extension_extract(to_extract, return_type="extensions")
ext_extract_all.keys()

# %%
str(ext_extract_all["Factor Inputs"])

# %% [markdown]
# Or merge the extracted data into a new pymrio Extension object (when passing a new name as return_type):

# %%
ext_new = mrio.extension_extract(to_extract, return_type="new_merged_extension")
str(ext_new)

# %% [markdown]
# CONT: Continue with explaining, mention the work with find_all etc

# %% [markdown]
# ### Search and extract

# %% [markdown]
# The extract methods can also be used in combination with the [search/explore](./explore.ipynb) methods of pymrio.
# This allows to search for specific rows and then extract the data.

# %% [markdown]
# For example, to extract all emissions from the air compartment we can use:

# %%
match_air = mrio.extension_match(find_all="air")

# %% [markdown]
# And then make a new extension object with the extracted data:

# %%
air_emissions = mrio.emissions.extract(match_air, return_type="extracted_air_emissions")
print(air_emissions)

# %% [markdown]
# For more information on the search methods see the [explore notebook](./explore.ipynb).



"""Utility function for pymrio.

KST 20140502
"""

import json
import logging
import os
import re
import ssl
import zipfile
from collections import namedtuple
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
import requests
import urllib3
from requests.adapters import HTTPAdapter

from pymrio.core.constants import DEFAULT_FILE_NAMES, LONG_VALUE_NAME, PYMRIO_PATH


def is_vector(inp):
    """Return true if the input can be interpreted as a 'true' vector.

    Note:
    ----
    Does only check dimensions, not if type is numeric

    Parameters
    ----------
    inp : numpy.ndarray or something that can be converted into ndarray

    Returns
    -------
    Boolean
        True for vectors: ndim = 1 or ndim = 2 and shape of one axis = 1
        False for all other arrays
    """
    inp = np.asarray(inp)
    nr_dim = np.ndim(inp)
    if nr_dim == 1 or (nr_dim == 2) and (1 in inp.shape):
        return True
    return False


def get_repo_content(path):
    """List of files in a repo (path or zip).

    Parameters
    ----------
    path: string or pathlib.Path

    Returns
    -------
    Returns a namedtuple with .iszip and .filelist
    The path in filelist are pure strings.

    """
    path = Path(path)

    if zipfile.is_zipfile(str(path)):
        with zipfile.ZipFile(str(path)) as zz:
            filelist = [info.filename for info in zz.infolist()]
        iszip = True
    else:
        iszip = False
        filelist = [str(f) for f in path.glob("**/*") if f.is_file()]
    return namedtuple("repocontent", ["iszip", "filelist"])(iszip, filelist)


def get_file_para(path, path_in_arc=""):
    """Read the file parameter file.

    Helper function to consistently read the file parameter file, which can
    either be uncompressed or included in a zip archive.  By default, the file
    name is to be expected as set in DEFAULT_FILE_NAMES['filepara'] (currently
    file_parameters.json), but can defined otherwise by including the file
    name of the parameter file in the parameter path.

    Parameters
    ----------
    path: pathlib.Path or string
        Path or path with para file name for the data to load.
        This must either point to the directory containing the uncompressed
        data or the location of a compressed zip file with the data. In the
        later case the parameter 'path_in_arc' needs to be specific to
        further indicate the location of the data in the compressed file.

    path_in_arc: string, optional
        Path to the data in the zip file (where the fileparameters file is
        located). path_in_arc must be given without leading dot and slash;
        thus to point to the data in the root of the compressed file pass ''
        (default), for data in e.g. the folder 'emissions' pass 'emissions/'.
        Only used if parameter 'path' points to an compressed zip file.

    Returns
    -------
    Returns a namedtuple with
    .folder: str with the absolute path containing the
           file parameter file. In case of a zip the path
           is relative to the root in the zip
    .name: Filename without folder of the used parameter file.
    .content: Dictionary with the content oft the file parameter file

    Raises
    ------
    FileNotFoundError if parameter file not found

    """
    path = Path(path)

    if zipfile.is_zipfile(str(path)):
        para_file_folder = str(path_in_arc)
        with zipfile.ZipFile(file=str(path)) as zf:
            files = zf.namelist()
    else:
        para_file_folder = str(path)
        files = [str(f) for f in path.glob("**/*")]

    if para_file_folder not in files:
        if zipfile.is_zipfile(str(path)):
            # b/c in win os.path.join adds \ also for within zipfile
            if para_file_folder != "":
                para_file_full_path = (para_file_folder + "/" + DEFAULT_FILE_NAMES["filepara"]).replace("//", "/")
            else:
                para_file_full_path = DEFAULT_FILE_NAMES["filepara"]

        else:
            para_file_full_path = os.path.join(para_file_folder, DEFAULT_FILE_NAMES["filepara"])
    else:
        para_file_full_path = para_file_folder
        para_file_folder = os.path.dirname(para_file_full_path)

    if para_file_full_path not in files:
        raise FileNotFoundError(f"File parameter file {para_file_full_path} not found")

    if zipfile.is_zipfile(str(path)):
        with zipfile.ZipFile(file=str(path)) as zf:
            para_file_content = json.loads(zf.read(para_file_full_path).decode("utf-8"))
    else:
        with open(para_file_full_path) as pf:
            para_file_content = json.load(pf)

    return namedtuple("file_parameter", ["folder", "name", "content"])(
        para_file_folder,
        os.path.basename(para_file_full_path),
        para_file_content,
    )


def build_agg_matrix(agg_vector, pos_dict=None):
    """Aggregate based on mapping given in input as numerical or str vector.

    Parameters
    ----------
        agg_vector : list or vector like numpy ndarray
            This can be row or column vector.
            Length m with position given for n and -1 if values
            should not be included
            or
            length m with id_string for the aggregation

        pos_dict : dictionary
            (only possible if agg_vector is given as string)
            output order for the new matrix
            must be given as dict with
            'string in agg_vector' = pos
            (as int, -1 if value should not be included in the aggregation)

    Returns
    -------
    agg_matrix : numpy ndarray
        Aggregation matrix with shape (n, m) with n (rows) indicating the
        new classification and m (columns) the old classification

    Examples
    --------
    Assume an input vector with either

    >>> inp1 = np.array([0, 1, 1, 2])

    or

    >>> inp2 = ['a', 'b', 'b', 'c']

    inp1 and inp2 are equivalent, the entries are just indicators.

    >>> build_agg_matrix(inp1)
    >>> build_agg_matrix(inp2)

    results in
    >>> array([[1., 0., 0., 0.],
    >>>        [0., 1., 1., 0.],
    >>>        [0., 0., 0., 1.]])

    The order can be defined by a dictionary, thus

    >>> pymrio.build_agg_matrix(np.array([1, 0, 0, 2]))

    is equivalent to

    >>> pymrio.build_agg_matrix(['b', 'a', 'a', 'c'], dict(a=0,b=1,c=2))

    >>> array([[0., 1., 1., 0.],
    >>>        [1., 0., 0., 0.],
    >>>        [0., 0., 0., 1.]])

    """
    if isinstance(agg_vector, np.ndarray):
        agg_vector = agg_vector.flatten().tolist()

    if type(list(agg_vector)[0]) is str:
        str_vector = agg_vector
        agg_vector = np.zeros(len(str_vector))
        if pos_dict:
            if len(pos_dict.keys()) != len(set(str_vector)):
                raise ValueError("Posistion elements inconsistent with aggregation vector")
            seen = pos_dict
        else:
            seen = {}
        counter = 0
        for ind, item in enumerate(str_vector):
            if item not in seen:
                seen[item] = counter
                counter += 1
            agg_vector[ind] = seen[item]

    agg_vector = np.array(agg_vector, dtype=int)
    agg_vector = agg_vector.reshape((1, -1))
    row_corr = agg_vector
    col_corr = np.arange(agg_vector.size)

    agg_matrix = np.zeros((row_corr.max() + 1, col_corr.max() + 1))
    agg_matrix[row_corr, col_corr] = 1

    # set columns with -1 value to 0
    agg_matrix[np.tile(agg_vector == -1, (np.shape(agg_matrix)[0], 1))] = 0

    return agg_matrix


def diagonalize_columns_to_sectors(df: pd.DataFrame, sector_index_level: Union[str, int] = "sector") -> pd.DataFrame:
    """Add the resolution of the rows to columns by diagonalizing.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to diagonalize
    sector_index_name : string, optional
        Name or number of the index level containing sectors.

    Returns
    -------
    pd.DataFrame, diagonalized


    Example:
    --------
        input       output
         (all letters are index or header)
            A B     A A A B B B
                    x y z x y z
        A x 3 1     3 0 0 1 0 0
        A y 4 2     0 4 0 0 2 0
        A z 5 3     0 0 5 0 0 3
        B x 6 9     6 0 0 9 0 0
        B y 7 6     0 7 0 0 6 0
        B z 8 4     0 0 8 0 0 4

    """
    sectors = df.index.get_level_values(sector_index_level).unique()
    sector_name = sector_index_level if type(sector_index_level) is str else "sector"

    diag_df = pd.DataFrame(
        data=diagonalize_blocks(df.values, blocksize=len(sectors)),
        index=df.index,
        columns=pd.MultiIndex.from_product([df.columns, sectors], names=[*df.columns.names, sector_name]),
    )
    return diag_df


def diagonalize_blocks(arr, blocksize: int):
    """Diagonalize sections of columns of an array for the whole array.

    Parameters
    ----------
    arr : numpy array
        Input array

    blocksize : int
        number of rows/colums forming one block

    Returns
    -------
    numpy ndarray with shape (columns 'arr' * blocksize,
                              columns 'arr' * blocksize)

    Example:
    --------
    arr:      output: (blocksize = 3)
        3 1     3 0 0 1 0 0
        4 2     0 4 0 0 2 0
        5 3     0 0 5 0 0 3
        6 9     6 0 0 9 0 0
        7 6     0 7 0 0 6 0
        8 4     0 0 8 0 0 4
    """
    nr_col = arr.shape[1]
    nr_row = arr.shape[0]

    if np.mod(nr_row, blocksize):
        raise ValueError("Number of rows of input array must be a multiple of blocksize")

    arr_diag = np.zeros((nr_row, blocksize * nr_col))

    for col_ind, col_val in enumerate(arr.T):
        col_start = col_ind * blocksize
        col_end = blocksize + col_ind * blocksize
        for _ind in range(int(nr_row / blocksize)):
            row_start = _ind * blocksize
            row_end = blocksize + _ind * blocksize
            arr_diag[row_start:row_end, col_start:col_end] = np.diag(col_val[row_start:row_end])

    return arr_diag


def set_dom_block(df: pd.DataFrame, value: float = 0) -> pd.DataFrame:
    """Set domestic blocks to value (0 by default).

    Requires that the region is the top index in the multiindex
    hierarchy (default case in pymrio).

    Parameters
    ----------
    df : pd.DataFrame
    value : float, optional

    Returns
    -------
    pd.DataFrame

    """
    regions = df.index.get_level_values(0).unique()
    df_res = df.copy()
    for reg in regions:
        df_res.loc[reg, reg] = value
    return df_res


def set_block(arr, arr_block):
    """Set the diagonal blocks of an array to an given array.

    Parameters
    ----------
    arr : numpy ndarray
        the original array
    block_arr : numpy ndarray
        the block array for the new diagonal

    Returns
    -------
    numpy ndarray (the modified array)

    """
    nr_col = arr.shape[1]
    nr_row = arr.shape[0]

    nr_col_block = arr_block.shape[1]
    nr_row_block = arr_block.shape[0]

    if np.mod(nr_row, nr_row_block) or np.mod(nr_col, nr_col_block):
        raise ValueError("Number of rows/columns of the input array must be a multiple of block shape")
    if nr_row / nr_row_block != nr_col / nr_col_block:
        raise ValueError("Block array can not be filled as diagonal blocks in the given array")

    arr_out = arr.copy()

    for row_ind in range(int(nr_row / nr_row_block)):
        row_start = row_ind * nr_row_block
        row_end = nr_row_block + nr_row_block * row_ind
        col_start = row_ind * nr_col_block
        col_end = nr_col_block + nr_col_block * row_ind

        arr_out[row_start:row_end, col_start:col_end] = arr_block

    return arr_out


def unique_element(ll):
    """Return unique elements from a list preserving the original order."""
    seen = {}
    result = []
    for item in ll:
        if item in seen:
            continue
        seen[item] = 1
        result.append(item)
    return result


def build_agg_vec(agg_vec, **source):
    """Build an combined aggregation vector considering diff classifications.

    This function build an aggregation vector based on the order in agg_vec.
    The naming and actual mapping is given in source, either explicitly or by
    pointing to a folder with the mapping.


    Parameters
    ----------
    agg_vec : list
        A list of sector or regions to which the IOSystem shall be aggregated.
        The order in agg_vec is important:
        If a string was assigned to one specific entry it will not be
        overwritten if it is given in the next vector, e.g.  ['EU', 'OECD']
        would aggregate first into EU and the remaining one into OECD, whereas
        ['OECD', 'EU'] would first aggregate all countries into OECD and than
        the remaining countries into EU.

    source : list or string
        Definition of the vectors in agg_vec.  The input vectors (either in the
        file or given as list for the entries in agg_vec) must be as long as
        the desired output with a string for every position which should be
        aggregated and None for position which should not be used.

        Special keywords:

            - path : Path to a folder with concordance matrices.
                     The files in the folder can have any extension but must be
                     in text format (tab separated) with one entry per row.
                     The last column in the file will be taken as aggregation
                     vectors (other columns can be used for documentation).
                     Values must be given for every entry in the original
                     classification (string None for all values not used) If
                     the same entry is given in source and as text file in
                     path than the one in source will be used.

                     Two special path entries are available so far:

                     - 'exio2'
                       Concordance matrices for EXIOBASE 2.0
                     - 'test'
                       Concordance matrices for the test IO system

                     If a entry is not found in source and no path is given
                     the current directory will be searched for the definition.

            - miss : Entry to use for missing values, default: 'REST'

    Returns
    -------
    list
        The aggregation vector

    Examples
    --------
    >>> build_agg_vec(['EU', 'OECD'], path = 'test')
    ['EU', 'EU', 'EU', 'OECD', 'REST', 'REST']

    >>> build_agg_vec(['OECD', 'EU'], path = 'test', miss='RoW')
    ['OECD', 'EU', 'OECD', 'OECD', 'RoW', 'RoW']

    >>> build_agg_vec(['EU', 'orig_regions'], path = 'test')
    ['EU', 'EU', 'EU', 'reg4', 'reg5', 'reg6']

    >>> build_agg_vec(['supreg1', 'other'], path = 'test',
    >>>        other = [None, None, 'other1', 'other1', 'other2', 'other2'])
    ['supreg1', 'supreg1', 'other1', 'other1', 'other2', 'other2']



    """
    # build a dict with aggregation vectors in source and folder
    # TODO: the logic here should be moved to constants
    if type(agg_vec) is str:
        agg_vec = [agg_vec]
    agg_dict = {}
    for entry in agg_vec:
        try:
            agg_dict[entry] = source[entry]
        except KeyError:
            folder = source.get("path", "./")
            folder = os.path.join(PYMRIO_PATH[folder], "concordance")
            for file in os.listdir(folder):
                if entry == os.path.splitext(file)[0]:
                    _tmp = np.genfromtxt(os.path.join(folder, file), dtype=str)
                    if _tmp.ndim == 1:
                        agg_dict[entry] = [None if ee == "None" else ee for ee in _tmp.tolist()]
                    else:
                        agg_dict[entry] = [None if ee == "None" else ee for ee in _tmp[:, -1].tolist()]
                    break
            else:
                logging.error(f"Aggregation vector -- {str(entry)} -- not found")

    # build the summary aggregation vector
    def _rep(ll, ii, vv):
        ll[ii] = vv

    miss_val = source.get("miss", "REST")

    vec_list = [agg_dict[ee] for ee in agg_vec]
    out = [
        None,
    ] * len(vec_list[0])
    for currvec in vec_list:
        if len(currvec) != len(out):
            logging.warning("Inconsistent vector length")
        [_rep(out, ind, val) for ind, val in enumerate(currvec) if not out[ind]]

    [_rep(out, ind, miss_val) for ind, val in enumerate(out) if not val]

    return out


def find_first_number(ll):
    """Return nr of first entry parseable to float in ll, None otherwise."""
    for nr, entry in enumerate(ll):
        try:
            float(entry)
        except (ValueError, TypeError):
            pass
        else:
            return nr
    return None


def sniff_csv_format(
    csv_file,
    potential_sep=None,
    max_test_lines=10,
    zip_file=None,
):
    r"""Attempt to get the separator, nr of index cols and header rows in a csv file.

    Parameters
    ----------
    csv_file: str
        Path to a csv file

    potential_sep: list, optional
        List of potential separators (delimiters) to test.
        Default: '\t', ',', ';', '|', '-', '_' (used when passing None)

    max_test_lines: int, optional
        How many lines to test, default: 10 or available lines in csv_file

    zip_file: str, optional
        Path to a zip file containing the csv file (if any, default: None).
        If a zip file is given, the path given at 'csv_file' is assumed
        to be the path to the file within the zip_file.

    Returns
    -------
        dict with
            sep: string (separator)
            nr_index_col: int
            nr_header_row: int

        Entries are set to None if inconsistent information in the file
    """
    potential_sep = potential_sep if potential_sep else ["\t", ",", ";", "|", "-", "_"]

    def read_first_lines(filehandle):
        lines = []
        for _ in range(max_test_lines):
            line = filehandle.readline()
            if line == "":
                continue
            try:
                line = line.decode("utf-8")
            except AttributeError:
                pass
            lines.append(line[:-1])
        return lines

    if zip_file:
        with zipfile.ZipFile(zip_file, "r") as zz:
            with zz.open(csv_file, "r") as ff:
                test_lines = read_first_lines(ff)
    else:
        with open(csv_file) as ff:
            test_lines = read_first_lines(ff)

    sep_aly_lines = [
        sorted(
            [(line.count(sep), sep) for sep in potential_sep if line.count(sep) > 0],
            key=lambda x: x[0],
            reverse=True,
        )
        for line in test_lines
    ]

    sep = None
    for nr, (count, sep) in enumerate(sep_aly_lines[0]):
        for line in sep_aly_lines:
            if line[nr][0] == count:
                break
        else:
            sep = None

        if sep:
            break

    lines_with_sep = [line for line in test_lines if sep in line]

    nr_header_row = None
    nr_index_col = None

    if sep:
        nr_index_col = find_first_number(lines_with_sep[-1].split(sep))
        if nr_index_col:
            for header_row, line in enumerate(lines_with_sep):
                nr_header_row = header_row
                if find_first_number(line.split(sep)) == nr_index_col:
                    break

    return {"sep": sep, "nr_header_row": nr_header_row, "nr_index_col": nr_index_col}


def filename_from_url(url):
    """Extract a file name from the download link for that file.

    Parameters
    ----------
    url: str,
        The download link of the file

    Returns
    -------
    str,
        The extracted file name

    """
    name = re.search(r"[^/\\&?]+\.\w{2,7}(?=([?&].*$|$))", url)
    if name:
        return name.group()
    raise ValueError("Could not extract filename from url")


def check_if_long(df, value_name=LONG_VALUE_NAME):
    """Check if a given DataFrame follows is in a long format.

    Currently this only checks if 'value_name' is in the columns.
    In no parameter is given, it uses the default value 'value', defined
    in constants.py

    Note:
    -----
    This function does not check if the DataFrame is actually in a long format,
    it only checks if the DataFrame has a column with the name 'value_name'.
    There is an edge case of having 'value_name' as a column name, but not
    being in a long format. This function will return True in this case.

    Returns
    -------
    bool
        True if the DataFrame is in a long format, False otherwise
    """
    if not isinstance(df, pd.DataFrame) or isinstance(df, pd.Series):
        return False

    # TODO: refactor once we have these in the constants.py
    if isinstance(df, pd.DataFrame):
        if "region" in df.columns.names:
            return False
        if "sector" in df.columns.names:
            return False
        if "category" in df.columns.names:
            return False

    if value_name not in df.columns:
        return False
    return True


def to_long(df, value_name=LONG_VALUE_NAME):
    """Convert the pymrio matrix df format to a long format.

    FIX: All index and columns become separate columns (not index!)

    Parameters
    ----------
    df: pymrio matrix
        The pymrio matrix (e.g. Z, F, etc) to convert

    value_name: str, optional
        The name of the value column, default: 'value'
        as defined in constants.py

    Returns
    -------
    pd.DataFrame
    """
    df_long = df.stack(df.columns.names)
    df_long.name = value_name
    df_long = pd.DataFrame(df_long)
    return df_long


def ssl_fix(*args, **kwargs):
    """Try to use a request connection with Lagacy option when normal connection fails.

    Parameters
    ----------
    Parameters of a normal requests.get() function
        url: URL for the new :class:`Request` object.
        params: (optional) Dictionary, list of tuples or bytes to send
        in the query string for the class `Request`.
        **kwargs: Optional arguments that `request` takes.

    Returns
    -------
        r: class:`Response <Response>` object
    """

    class CustomHttpAdapter(HTTPAdapter):
        # "Transport adapter" that allows us to use custom ssl_context.

        def __init__(self, ssl_context=None, **kwargs):
            self.ssl_context = ssl_context
            super().__init__(**kwargs)

        def init_poolmanager(self, connections, maxsize, block=False):
            self.poolmanager = urllib3.poolmanager.PoolManager(
                num_pools=connections,
                maxsize=maxsize,
                block=block,
                ssl_context=self.ssl_context,
            )

    try:
        r = requests.get(*args, **kwargs)
    except Exception:
        ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        ctx.options |= 0x4  # OP_LEGACY_SERVER_CONNECT
        session = requests.session()
        session.mount("https://", CustomHttpAdapter(ctx))
        r = session.get(*args, **kwargs)

    return r


def index_fullmatch(df_ix, find_all=None, **kwargs):
    """Fullmatch regex on index of df_ix.

    Similar to pandas str.fullmatch, thus the whole
    string of the index must match.

    The index levels need to be named (df.index.name needs to
    be set for all levels).

    Note:
    -----
    Arguments are set to case=True, flags=0, na=False, regex=True.
    For case insensitive matching, use (?i) at the beginning of the pattern.

    See the pandas/python.re documentation for more details.


    Parameters
    ----------
    df_ix : pd.DataFrame, pd.Series, pd.Index or pd.MultiIndex
        Method works on Index directly or extract from DataFrame/Series
    find_all : None or str
        If str (regex pattern) search for all matches in all index levels.
        All matching rows are returned. The remaining kwargs are ignored.
    kwargs : dict
        The regex to match. The keys are the index names,
        the values are the regex to match.
        If the entry is not in index name, it is ignored silently.

    Returns
    -------
    pd.DataFrame, pd.Series, pd.Index or pd.MultiIndex
        The matched rows/index, same type as _dfs_idx

    """
    return _index_regex_matcher(_dfs_idx=df_ix, _method="fullmatch", _find_all=find_all, **kwargs)


def index_match(df_ix, find_all=None, **kwargs):
    """Match regex on index of df_ix.

    Similar to pandas str.match, thus the start of the index string must match.

    The index levels need to be named (df.index.name needs to
    be set for all levels).

    Note:
    -----
    Arguments are set to case=True, flags=0, na=False, regex=True.
    For case insensitive matching, use (?i) at the beginning of the pattern.

    See the pandas/python.re documentation for more details.


    Parameters
    ----------
    df_ix : pd.DataFrame, pd.Series, pd.Index or pd.MultiIndex
        Method works on Index directly or extract from DataFrame/Series
    find_all : None or str
        If str (regex pattern) search for all matches in all index levels.
        All matching rows are returned. The remaining kwargs are ignored.
    kwargs : dict
        The regex to match. The keys are the index names,
        the values are the regex to match.
        If the entry is not in index name, it is ignored silently.

    Returns
    -------
    pd.DataFrame, pd.Series, pd.Index or pd.MultiIndex
        The matched rows/index, same type as _dfs_idx

    """
    return _index_regex_matcher(_dfs_idx=df_ix, _method="match", _find_all=find_all, **kwargs)


def index_contains(df_ix, find_all=None, **kwargs):
    """Check if index contains a regex pattern.

    Similar to pandas str.contains, thus the index
    string must contain the regex pattern.

    The index levels need to be named (df.index.name needs to
    be set for all levels).

    Note:
    -----
    Arguments are set to case=True, flags=0, na=False, regex=True.
    For case insensitive matching, use (?i) at the beginning of the pattern.

    See the pandas/python.re documentation for more details.


    Parameters
    ----------
    df_ix : pd.DataFrame, pd.Series, pd.Index or pd.MultiIndex
        Method works on Index directly or extract from DataFrame/Series
    find_all : None or str
        If str (regex pattern) search for all matches in all index levels.
        All matching rows are returned. The remaining kwargs are ignored.
    kwargs : dict
        The regex to match. The keys are the index names,
        the values are the regex to match.
        If the entry is not in index name, it is ignored silently.

    Returns
    -------
    pd.DataFrame, pd.Series, pd.Index or pd.MultiIndex
        The matched rows/index, same type as _dfs_idx

    """
    return _index_regex_matcher(_dfs_idx=df_ix, _method="contains", _find_all=find_all, **kwargs)


def _index_regex_matcher(_dfs_idx, _method, _find_all=None, **kwargs):
    """Match index of df with regex.

    The generic method for the contain, match, fullmatch implementation
    along the index of the pymrio dataframes.

    The index levels need to be named (df.index.name needs to
    be set for all levels).

    Note:
    -----
    Arguments are set to case=True, flags=0, na=False, regex=True.
    For case insensitive matching, use (?i) at the beginning of the pattern.

    See the pandas/python.re documentation for more details.


    Parameters
    ----------
    _dfs_idx : pd.DataFrame, pd.Series, pd.Index or pd.MultiIndex
        Method works on Index directly or extract from DataFrame/Series
    _method : str
        The method to use for matching, one of 'contains', 'match', 'fullmatch'
    _find_all : None or str
        If str (regex pattern) search for all matches in all index levels.
        All matching rows are returned. The remaining kwargs are ignored.
    kwargs : dict
        The regex to match. The keys are the index names,
        the values are the regex to match.
        If the entry is not in index name, it is ignored silently.

    Returns
    -------
    pd.DataFrame, pd.Series, pd.Index or pd.MultiIndex
        The matched rows/index, same type as _dfs_idx

    """
    if _method not in ["contains", "match", "fullmatch"]:
        raise ValueError('Method must be one of "contains", "match", "fullmatch"')

    if _find_all is not None:
        if type(_dfs_idx) in [pd.DataFrame, pd.Series]:
            idx = _dfs_idx.index
        elif type(_dfs_idx) in [pd.Index, pd.MultiIndex]:
            idx = _dfs_idx
        else:
            raise ValueError("Type of _dfs_idx must be one of pd.DataFrame, pd.Series, pd.Index or pd.MultiIndex")
        found = np.array([], dtype=int)
        for idx_name in idx.names:
            fun = getattr(idx.get_level_values(idx_name).str, _method)
            ff = idx[fun(_find_all, case=True, flags=0, na=False)]
            found = np.append(found, idx.get_indexer(ff))
        found = np.unique(found)
        if type(_dfs_idx) in [pd.DataFrame, pd.Series]:
            return _dfs_idx.iloc[found]
        return _dfs_idx[found]

    at_least_one_valid = False
    for key, value in kwargs.items():
        try:
            if type(_dfs_idx) in [pd.DataFrame, pd.Series]:
                fun = getattr(_dfs_idx.index.get_level_values(key).str, _method)
            elif type(_dfs_idx) in [pd.Index, pd.MultiIndex]:
                fun = getattr(_dfs_idx.get_level_values(key).str, _method)
            else:
                raise ValueError("Type of _dfs_idx must be one of pd.DataFrame, pd.Series, pd.Index or pd.MultiIndex")
            _dfs_idx = _dfs_idx[fun(value, case=True, flags=0, na=False)]
            at_least_one_valid = True
        except KeyError:
            pass

    if not at_least_one_valid:
        if type(_dfs_idx) in [pd.DataFrame, pd.Series]:
            _dfs_idx = pd.DataFrame(columns=_dfs_idx.columns)
        elif type(_dfs_idx) in [pd.Index, pd.MultiIndex]:
            _dfs_idx = pd.Index([])

    return _dfs_idx


def _characterize_get_requried_col(
    factors,
    ext_index_names,
    characterized_name_column,
    characterization_factors_column,
    characterized_unit_column,
):
    """Check and return a list of required columns.

    For paramters naming see function characterize.

    """
    req_index = ext_index_names.copy()

    if "region" in factors.columns:
        req_index.append("region")
    if "sector" in factors.columns:
        req_index.append("sector")

    required_columns = req_index + [
        characterization_factors_column,
        characterized_name_column,
        characterized_unit_column,
    ]

    if not set(required_columns).issubset(set(factors.columns)):
        raise ValueError("Not all required columns in the passed DataFrame >factors<")

    return namedtuple("ret_val", ["required_index_col", "all_required_columns"])(req_index, required_columns)


def _validate_characterization_table(
    factors,
    all_required_col,
    regions,
    sectors,
    ext_unit,
    characterized_name_column="impact",
    characterized_unit_column="impact_unit",
    orig_unit_column="stressor_unit",
):
    """Check a factors sheet for characterization (internal function).

    This should not be called directly, use characterize with only_validation instead.

    Given a factors sheet reports error in

        - unit errors (impact unit consistent, stressor unit match).
            Note: does not check if the conversion is correct!
        - report missing stressors, regions, sectors which are in factors
            but not in the extension
        - if factors are specified for all regions/sectors of the extension

    Besides the unit errors, factors can still be passed to the characterization
    routine, and missing data will be assumed to be 0.

    Parameters follow the convention of the characterization method:

    Parameters
    ----------
    factors: pd.DataFrame
        A dataframe in long format with numerical index and columns named
        index.names of the extension to be characterized and
        'characterized_name_column', 'characterization_factors_column',
        'characterized_unit_column', 'orig_unit_column'

    regions: Index
        Regions in the mrio object (from get_regions())

    sectors: Index
        Sectors in the mrio object (from get_sectors())

    ext_unit: pd.DataFrame
        The mrio.unit dataframe

    characterized_name_column: str (optional)
        Name of the column with the names of the
        characterized account (default: "impact")

    characterization_factors_column: str (optional)
        Name of the column with the factors for the
        characterization (default: "factor")

    characterized_unit_column: str (optional)
        Name of the column with the units of the characterized accounts
        characterization (default: "impact_unit")

    Returns
    -------
    pd.DataFrame: factors sheet with additional error columns

    """
    # searching the next available df to get the
    # index of stressors/regions/sector depending on input

    fac = factors.copy()

    fac.loc[:, "error_unit_impact"] = False
    fac.loc[:, "error_unit_stressor"] = False
    fac.loc[:, "error_missing_stressor"] = False

    if "region" in fac.columns:
        fac.loc[:, "error_missing_region"] = False
    if "sector" in fac.columns:
        fac.loc[:, "error_missing_sector"] = False

    unique_impacts = fac.loc[:, characterized_name_column].unique()

    # Check for consistent units per impact
    for imp in unique_impacts:
        if not (
            fac.loc[
                fac.loc[:, characterized_name_column] == imp,
                characterized_unit_column,
            ]
            .eq(
                fac.loc[
                    fac.loc[:, characterized_name_column] == imp,
                    characterized_unit_column,
                ].iloc[0]
            )
            .all()
        ):
            fac.loc[
                fac.loc[:, characterized_name_column] == imp,
                "error_unit_impact",
            ] = True

    # unit per stressor check
    fac = fac.set_index(ext_unit.index.names).sort_index()
    if "unit" in fac.columns:
        um = fac.join(ext_unit, how="inner", rsuffix="_ext_orig")
        ext_unit_col = "unit_ext_orig"
    else:
        um = fac.join(ext_unit, how="inner")
        ext_unit_col = "unit"
    fac = fac.reset_index()
    um = um.reset_index()

    fac.error_unit_stressor = fac.error_unit_stressor.astype("object")
    fac.loc[:, "error_unit_stressor"] = um.loc[:, orig_unit_column] != um.loc[:, ext_unit_col]
    with pd.option_context("future.no_silent_downcasting", True):
        fac.loc[:, "error_unit_stressor"] = fac.loc[:, "error_unit_stressor"].fillna(False)
    fac.error_unit_stressor = fac.error_unit_stressor.infer_objects(copy=False)

    fac = fac.set_index(ext_unit.index.names).sort_index()
    for row in fac.index.unique():
        # Stressor not in the extension
        if row not in ext_unit.index:
            fac.loc[row, "error_missing_stressor"] = True
            continue

        # Checking for region/sector data
        # This covers if not all regions present in the mrio
        # are specified in the factors sheet.
        # In that case, the full error_missing_region/sector for the
        # the full region is set to True
        if "region" in all_required_col:
            reg_cov = (
                fac.loc[row, ["region", characterized_name_column]].groupby(characterized_name_column).region.apply(set)
            )
            for improw in reg_cov.index:
                if len(regions.difference(reg_cov[improw])) > 0:
                    fac.loc[row, "error_missing_region"] = True

        if "sector" in all_required_col:
            reg_cov = (
                fac.loc[row, ["sector", characterized_name_column]].groupby(characterized_name_column).sector.apply(set)
            )
            for improw in reg_cov.index:
                if len(sectors.difference(reg_cov[improw])) > 0:
                    fac.loc[row, "error_missing_sector"] = True

    # check if additional region/sectors in the data
    if "region" in all_required_col:
        for reg in fac.region.unique():
            if reg not in regions:
                fac.loc[fac.region == reg, "error_missing_region"] = True
    if "sector" in all_required_col:
        for sec in fac.sector.unique():
            if sec not in sectors:
                fac.loc[fac.sector == sec, "error_missing_sector"] = True

    return fac.reset_index()


def extend_rows(df, **kwargs):
    """Given a df, duplicate rows by spreading one columns.

    This function takes a dataframe (e.g. a bridge or characterization specs),
    and spreads rows based on the input given in the keyword arguments.

    Each keyword can be a column header, with the argument being a dict with the
    values to be spread.

    Example:
    -------
    >>> df = pd.DataFrame(
    ...     {
    ...         "region": ["GLO", "GLO", "GLO"],
    ...         "sector": ["A", "B", "C"],
    ...         "value": [1, 2, 3],
    ...     }
    ... )
    >>> extend_rows(df, region={"GLO": ["reg1", "reg2", "reg3"]},
    ...             sector={"B": ["b1", "b2"]})
       region sector  value
    0    reg1      A      1
    1    reg1      C      3
    2    reg2      A      1
    3    reg2      C      3
    4    reg3      A      1
    5    reg3      C      3
    6    reg1     b1      2
    7    reg2     b1      2
    8    reg3     b1      2
    9    reg1     b2      2
    10   reg2     b2      2
    11   reg3     b2      2

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame to extend with rows. Must have a numerical index.
    **kwargs : dict
        Column names as keys and dictionaries as values. Each dictionary maps
        original values in the column to lists of new values that will replace
        the original value in the new rows.

    Returns
    -------
    pandas.DataFrame
        A new DataFrame with rows spread according to the mapping in kwargs,
        sorted by the columns specified in kwargs.

    Raises
    ------
    ValueError
        If the DataFrame index is not a RangeIndex (numerical)
        If a specified column in kwargs is not in the DataFrame
        If no rows are found to spread for a value specified in the mapping


    """
    if not pd.api.types.is_numeric_dtype(df.index):
        raise ValueError("DataFrame index must be numerical")

    result = df.copy()

    for column, mapping in kwargs.items():
        if column not in result.columns:
            raise ValueError(f"Column {column} not in DataFrame")
        for original_value, new_values in mapping.items():
            new_dfs = []
            # Select rows with the original value in the specified column
            rows_to_spread = result[result[column] == original_value]
            if rows_to_spread.empty:
                raise ValueError(f"No rows found to spread for value: {original_value}")
            # Create new dataframes with the new values
            for new_value in new_values:
                new_df = rows_to_spread.copy()
                new_df[column] = new_value
                new_dfs.append(new_df)
            # Remove the original rows that have been spread
            if new_dfs:
                result = result[result[column] != original_value]
                result = pd.concat([result] + new_dfs, ignore_index=True)

    return result.sort_values(by=list(kwargs.keys()))


def check_df_map(df_orig, df_map):
    """Check which entries of df_map would be in effect given df_orig."""
    # TODO: we need a way to check for spelling mistakes
    # and other hickups sneaking into df_map.
    # In this function, we check for each line of df_map which entries
    # would be in effect given df_orig.
    pass


# CONT:
# TODO: bridge column indicator as argument (also update convert notebook)
# TODO: new default <->


def convert(
    df_orig,
    df_map,
    agg_func="sum",
    drop_not_bridged_index=True,
    ignore_columns=None,
    reindex=None,
):
    """Convert a DataFrame to a new classification.

    Parameters
    ----------
    df_orig : pd.DataFrame
        The DataFrame to process.
        The index/columns levels need to be named (df.index.name
        and df.columns.names needs to be set for all levels).
        All index to be bridged to new names need to be in the index (these are columns
        indicated with two underscores '__' in the mapping dataframe, df_map).
        Other constraining conditions (e.g. regions, sectors) can be either
        in the index or columns. If the same name exists in the
        index and columns, the values in index are preferred.

    df_map : pd.DataFrame
        The DataFrame with the mapping of the old to the new classification.
        This requires a specific structure, which depends on the structure of the
        dataframe to be characterized:

        - Constraining data (e.g. stressors, regions, sectors) can be
          either in the index or columns of df_orig. The need to have the same
          name as the named index or column in df_orig. The algorithm searches
          for matching data in df_orig based on all constraining columns in df_map.

        - Bridge columns are columns with '__' in the name. These are used to
          map (bridge) some/all of the constraining columns in df_orig to the new
          classification.

        - One column "factor", which gives the multiplication factor for the
          conversion. If it is missing, it is set to 1.


        This is better explained with an example.
        Assuming a original dataframe df_orig with
        index names 'stressor' and 'compartment' and column name 'region',
        the characterizing dataframe could have the following structure (column names):

        stressor ... original index name
        compartment ... original index name
        region ... original column name
        factor ... the factor for multiplication/characterization
            If no factor is given, the factor is assumed to be 1.
            This can be used, to simplify renaming/aggregation mappings.
        impact__stressor ... the new index name,
            replacing the previous index name "stressor".
            Thus here "stressor" will be renamed to "impact", and the row index
            will be renamed by the entries here.
        compartment__compartment ... the new compartment,
            replacing the original compartment. No rename of column happens here,
            still row index will be renamed as given here.

        the columns with __ are called bridge columns, they are used
        to match the original index. The new dataframe with have index names
        based on the first part of the bridge column, in the order
        in which the bridge columns are given in the mapping dataframe.

        "region" is constraining column, these can either be for the index or column
        in df_orig. In case both exist, the one in index is preferred.


    agg_func : str or func
        the aggregation function to use for multiple matchings (summation by default)

    drop_not_bridged_index : bool, optional
        What to do with index levels in df_orig not appearing in the bridge columns.
        If True, drop them after aggregation across these, if False,
        pass them through to the result.

        *Note:* Only index levels will be dropped, not columns.

        In case some index levels need to be dropped, and some not
        make a bridge column for the ones to be dropped and map all to the same name.
        Then drop this index level after the conversion.

    ignore_columns : list, optional
        List of column names in df_map which should be ignored.
        These could be columns with additional information, unit columns, etc.

    reindex: str, None or collection
        Wrapper for pandas' reindex method to control return order.
        - If None: sorts the index alphabetically.
        - If str: uses the unique value order from the bridge column as the index order.
        - For other types (e.g., collections): passes directly to pandas.reindex.

    """
    bridge_columns = [col for col in df_map.columns if "__" in col]

    # groupby breaks with NaNs or None, fix it here
    df_map.loc[:, bridge_columns] = df_map.loc[:, bridge_columns].fillna("")

    unique_new_index = df_map.loc[:, bridge_columns].drop_duplicates().set_index(bridge_columns).index

    bridge_components = namedtuple("bridge_components", ["new", "orig", "raw"])
    bridges = []

    if not ignore_columns:
        ignore_columns = []

    if isinstance(df_orig, pd.Series):
        df_orig = pd.DataFrame(df_orig)

    # some consistency checks of arguments and restructuring if everything is ok
    if len(bridge_columns) == 0:
        raise ValueError("No columns with '__' in the mapping DataFrame")
    for col in bridge_columns:
        if col.count("__") == 1:
            bridge = bridge_components(*col.split("__"), col)
        else:
            raise ValueError(f"Column {col} contains more then one '__'")
        if bridge.orig not in df_map.columns:
            raise ValueError(f"Column {bridge.orig} not in df_map")
        if (bridge.orig not in df_orig.index.names) and (bridge.orig not in df_orig.columns.names):
            raise ValueError(f"Column {bridge.orig} not in df_orig")
        bridges.append(bridge)

    orig_index_not_bridged = [ix for ix in df_orig.index.names if ix not in [b.orig for b in bridges]]

    df_map = df_map.set_index(bridge_columns)

    stacked_columns = []
    orig_column_index = df_orig.columns
    for col in df_map.columns:
        if col in (["factor"] + ignore_columns):
            continue
        if col not in df_orig.index.names:
            if col in df_orig.columns.names:
                df_orig = df_orig.stack(col, future_stack=True)
                stacked_columns.append(col)
                if isinstance(df_orig, pd.Series):
                    df_orig.name = "FOO_CONVERT_REMOVE"
                    df_orig = pd.DataFrame(df_orig)
            else:
                print("TODO: log warning for missing column")

    res_collector = []

    # loop over each new impact/characterized value
    # and collect entries, multiply and rename
    for entry in unique_new_index:
        df_cur_map = df_map.loc[[entry]]
        collector = []

        # the loop for getting all (potential regex) matches and multiplication
        for row in df_cur_map.iterrows():
            matched_entries = index_fullmatch(df_ix=df_orig, **row[1].to_dict())
            try:
                mul_entries = matched_entries * row[1].factor
            except AttributeError:
                mul_entries = matched_entries
            collector.append(mul_entries)

        df_collected = pd.concat(collector, axis=0)

        # renaming part, checks if the old name (bridge.orig) is in the current index
        # and renames by the new one (bridge.new)

        already_renamed = {}

        for bridge in bridges:
            # encountering a bridge with the same orig name but which should
            # lead to two new index levels
            if bridge.orig in already_renamed:
                # already renamed the index to another one previously,
                # but we need to create more index levels for the
                # same original index level
                new_index_value = df_cur_map.index.get_level_values(bridge.raw)[0]
                _old_index = df_collected.index.to_frame()
                # as we go along in order, we add them to the end of the index
                _old_index.insert(len(_old_index.columns), bridge.new, new_index_value)
                df_collected.index = pd.MultiIndex.from_frame(_old_index)

            else:
                for idx_old_names in df_collected.index.names:
                    if bridge.orig in idx_old_names:
                        # rename the index names
                        if isinstance(df_collected.index, pd.MultiIndex):
                            df_collected.index = df_collected.index.set_names(bridge.new, level=idx_old_names)
                        else:
                            df_collected.index = df_collected.index.set_names(bridge.new, level=None)

                        # rename the actual index values
                        df_collected = df_collected.reset_index(level=bridge.new)
                        for row in df_cur_map.reset_index().iterrows():
                            new_row_name = row[1][bridge.raw]
                            old_row_name = row[1][bridge.orig]
                            df_collected.loc[:, bridge.new] = df_collected.loc[:, bridge.new].str.replace(
                                pat=old_row_name, repl=new_row_name, regex=True
                            )

                        # put the index back
                        if df_collected.index.name is None:
                            # The case with a single index where the
                            # previous reset index
                            # left only a numerical index
                            df_collected = df_collected.set_index(bridge.new, drop=True, append=False)
                        else:
                            df_collected = df_collected.set_index(bridge.new, drop=True, append=True)

                        already_renamed[bridge.orig] = bridge

        res_collector.append(df_collected.groupby(by=df_collected.index.names).agg(agg_func))

    all_result = pd.concat(res_collector, axis=0)

    if len(stacked_columns) > 0:
        all_result = all_result.unstack(stacked_columns)
        try:
            all_result = all_result.loc[:, "FOO_CONVERT_REMOVE"]
        except KeyError:
            pass
        if len(orig_column_index.names) > 1:
            all_result = all_result.reorder_levels(orig_column_index.names, axis=1)
        all_result = all_result.reindex(columns=orig_column_index)

    if drop_not_bridged_index:
        all_result = all_result.reset_index(level=orig_index_not_bridged, drop=True)
    else:
        # move the not bridged index levels to the end of the index
        new_index = [ix for ix in all_result.index.names if ix not in orig_index_not_bridged]
        try:
            all_result = all_result.reorder_levels(new_index + orig_index_not_bridged)
        except TypeError:  # case where there is only one index level
            pass

    grouped = all_result.groupby(by=all_result.index.names).agg(agg_func)

    if reindex is not None:
        if isinstance(reindex, str):
            df_map = df_map.reset_index()
            if reindex in bridge_columns:
                grouped_order = grouped.reindex(index=df_map.loc[:, reindex].unique())
            else:
                raise ValueError(
                    f"Reindexing by {reindex} is not possible, it is not a bridge column in the mapping DataFrame."
                )
        else:
            grouped_order = grouped.reindex(index=reindex)
    else:
        grouped_order = grouped.sort_index()

    return grouped_order
